/*
 * qdvbts_pmt.cpp
 *
 *  Created on: May 17, 2013
 *      Author: root
 */

#include "../Includes/qstdinc.h"
#include "../Includes/qdvbts.h"
#include "qdvbts_priv.h"

QDVBPMT::QDVBPMT()
{
	Prepare(0, true);
}

QDVBPMT::~QDVBPMT()
{
	Prepare(0, true);
}

void QDVBPMT::Prepare(int nSID, bool bCurrent)
{
	m_nSID = nSID;					// Program ID (SID) to parse
	m_nPCR_PID = -1;

	m_bDoCurrent = bCurrent;		// Do currents or announces
	m_nCurVer = -1;					// Current version

	m_bCurrents = false;			// current_next_indicator N1
	m_bAnnounces = false;			// current_next_indicator N2

	/* Delete allocated desc and strm */
	for (int i=0; i<(int)m_descs.size(); i++)
	{
		QDVBDesc* pDesc = m_descs[i];
		delete pDesc;
	}

	for (int i=0; i<(int)m_streams.size(); i++)
	{
		QDVBStrm* pStrm = m_streams[i];
		delete pStrm;
	}

	m_descs.clear();
	m_streams.clear();


	PrepareChecks();
}

void QDVBPMT::RePrepare()
{
	Prepare(m_nSID, m_bDoCurrent);
}

void QDVBPMT::PrepareChecks()
{
	m_donesec[0] = m_donesec[1] = m_donesec[2] = m_donesec[3] = 0;
	m_checksec[0] = m_checksec[1] = m_checksec[2] = m_checksec[3] = 0;
}

void QDVBPMT::AddCheck(int nCurSec, int nLastSec)
{
	/* Prepare check sect */
	if ((m_checksec[0] | m_checksec[1] | m_checksec[2] | m_checksec[3]) == 0)
	{
		for (int i=0; i<=nLastSec; i++)
		{
			uint64_t nTmp = (uint64_t)i;
			uint64_t nPos = nTmp / 64;
			uint64_t nMask = 1;
			nMask = nMask << (nTmp % 64);
			m_checksec[nPos] |= nMask;
		}
	}

	/* Set bit in done sect */
	{
		uint64_t nTmp = (uint64_t)nCurSec;
		uint64_t nPos = nTmp / 64;
		uint64_t nMask = 1;
		nMask = nMask << (nTmp % 64);
		m_donesec[nPos] |= nMask;
	}
}
bool QDVBPMT::AddData(bool bStart, uint8_t* pData, int nSize)
{
	bool bRetVal = false;
	bool bAdded = false;
	if (bStart)
	{
		if (m_section.GetSize() == 0)
		{
			m_section.Add(pData, nSize);
			bAdded = true;
		}
	}
	else
	{
		if (m_section.GetSize())
		{
			m_section.Add(pData, nSize);
			bAdded = true;
		}
	}

	if (bAdded && (m_section.GetSize() > 0))
	{
		if (!ParseData(m_section.GetData(), (int)m_section.GetSize()))
		{
			m_section.Clean();
		}
		else
		{
			bRetVal = true;
		}
	}

	return bRetVal;
}

bool QDVBPMT::ParseData(uint8_t* pData, int nSize)
{
	QDVB_Start(pData, nSize);

	uint8_t nBytesToSkip = 0;
	uint16_t section_length = 0;
	uint16_t program_number = 0;
	uint8_t current_next_indicator = 0;
	uint8_t version_number = 0;
	uint8_t section_number = 0;
	uint8_t last_section_number = 0;
	int program_info_length = 0;
	uint8_t descriptor_length = 0;
	QDVBDesc* pDesc = NULL;
	QDVBStrm* pStrm = NULL;
	int ES_info_length = 0;
	bool bCRCOK = false;
	uint8_t* pCRCStartData = NULL;
	int nDiff = 0;

	QDVB_ReadBits(nBytesToSkip, 8);
	QDVB_SkipBits(nBytesToSkip * 8);

	pCRCStartData = QDVB_CurrentBuffer();

	QDVB_ReadBitsCheck(8, 2);		// PMT should have 2
	QDVB_ReadBitsCheck(1, 1);
	QDVB_ReadBitsCheck(1, 0);
	QDVB_SkipBits(2);

	QDVB_ReadBits(section_length, 12);
	if (section_length > 1021) QDVB_GotError();
	if (section_length > QDVB_CurrentSize()) return true;
	nDiff = QDVB_CurrentSize() - section_length;

	QDVB_ReadBits(program_number, 16);
	QDVB_SkipBits(2);

	QDVB_ReadBits(version_number, 5);
	QDVB_ReadBits(current_next_indicator, 1);
	if (current_next_indicator == 1) m_bCurrents = true; else m_bAnnounces = true;
	if (m_bDoCurrent && (current_next_indicator == 0)) QDVB_GotSuccess();
	if (!m_bDoCurrent && (current_next_indicator == 1)) QDVB_GotSuccess();

	QDVB_ReadBits(section_number, 8);
	QDVB_ReadBits(last_section_number, 8);

	QDVB_SkipBits(3);
	QDVB_ReadBits(m_nPCR_PID, 13);
	QDVB_SkipBits(4);
	QDVB_ReadBits(program_info_length, 12);
	if (program_info_length > 1021) QDVB_GotError();

	while (program_info_length > 0)
	{
		pDesc = new QDVBDesc();

		if (program_info_length <= 0) QDVB_GotError(); QDVB_ReadBits(pDesc->descriptor_tag, 8); program_info_length--;
		if (program_info_length <= 0) QDVB_GotError(); QDVB_ReadBits(descriptor_length, 8); program_info_length--;
		pDesc->data.resize(descriptor_length);

		/* Copy data */
		if (program_info_length < descriptor_length) QDVB_GotError();
		memcpy(pDesc->data.data(), QDVB_CurrentBuffer(), descriptor_length);
		program_info_length -= descriptor_length;
		QDVB_SkipBits(descriptor_length * 8);

		m_descs.push_back(pDesc);
		pDesc = NULL;
	}

	while ((((int)QDVB_CurrentSize()) - nDiff) > 4)
	{
		pStrm = new QDVBStrm();
		QDVB_ReadBits(pStrm->stream_type, 8);
		QDVB_SkipBits(3);
		QDVB_ReadBits(pStrm->elementary_PID, 13);
		QDVB_SkipBits(4);
		QDVB_ReadBits(ES_info_length, 12);

		if (ES_info_length > 1021) QDVB_GotError();

		// Read descs here
		while (ES_info_length > 0)
		{
			pDesc = new QDVBDesc();

			if (ES_info_length <= 0) QDVB_GotError(); QDVB_ReadBits(pDesc->descriptor_tag, 8); ES_info_length--;
			if (ES_info_length <= 0) QDVB_GotError(); QDVB_ReadBits(descriptor_length, 8); ES_info_length--;
			pDesc->data.resize(descriptor_length);

			/* Copy data */
			if (ES_info_length < descriptor_length) QDVB_GotError();
			memcpy(pDesc->data.data(), QDVB_CurrentBuffer(), descriptor_length);
			ES_info_length -= descriptor_length;
			QDVB_SkipBits(descriptor_length * 8);

			pStrm->descs.push_back(pDesc);
			pDesc = NULL;
		}
		m_streams.push_back(pStrm);
		pStrm = NULL;
	}

	if ((bCRCOK = QDVBValidCRC(pCRCStartData, section_length + 3)))
	{
		m_nCurVer = version_number;
		AddCheck(section_number, last_section_number);
	}
	else
	{
		QDVB_GotError();
	}

QDVB_BoundsError;

	if (pDesc) delete pDesc;
	if (pStrm) delete pStrm;
	return QDVB_BCError() && bCRCOK;
}

int QDVBPMT::GetSID()
{
	return m_nSID;
}

int QDVBPMT::GetPCR()
{
	return m_nPCR_PID;
}

void QDVBPMT::GetDescs(QDVBDescs& descs)
{
	for (int i=0; i<(int)m_descs.size(); i++)
	{
		descs.push_back(m_descs[i]->Duplicate());
	}
}

void QDVBPMT::GetStreams(QDVBStrms& streams)
{
	for (int i=0; i<(int)m_streams.size(); i++)
	{
		streams.push_back(m_streams[i]->Duplicate());
	}
}

bool QDVBPMT::HasCurrent()
{
	return m_bCurrents;
}

bool QDVBPMT::HasAnnounces()
{
	return m_bAnnounces;
}

int QDVBPMT::GetVersion()
{
	return m_nCurVer;
}

bool QDVBPMT::SecsReady()
{
	if ((m_checksec[0] | m_checksec[1] | m_checksec[2] | m_checksec[3]) == 0) return false;
	return (m_donesec[0] == m_checksec[0]) &&
			(m_donesec[1] == m_checksec[1]) &&
			(m_donesec[2] == m_checksec[2]) &&
			(m_donesec[3] == m_checksec[3]);
}

bool QDVBPMT::Ready()
{
	return SecsReady();
}
